# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import os

from future.builtins import str, int
import pytest
import logging
from astrophysix.simdm import SimulationStudy, Project, ProjectCategory
from astrophysix.simdm.protocol import AlgoType, Algorithm, InputParameter, PhysicalProcess, Physics
from astrophysix.simdm.experiment import Simulation, AppliedAlgorithm, ParameterSetting, ParameterVisibility,\
    ResolvedPhysicalProcess, PostProcessingRun
from astrophysix.simdm.protocol import SimulationCode, PostProcessingCode
from astrophysix.simdm.experiment.base import Experiment
from astrophysix.simdm.results import GenericResult, Snapshot
from astrophysix.simdm.datafiles import YamlFile, AsciiFile
from astrophysix.simdm.utils import DataType
from astrophysix import units as U


class TestAppliedAlgorithm(object):
    def test_applied_algo_init(self):
        """
        Tests AppliedAlgorithm instance initialisation
        """
        # Applied algorithm initialisation
        amr_algo = Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement)
        amr = AppliedAlgorithm(algorithm=amr_algo, details="Implementation #3")

        assert amr.implementation_details == "Implementation #3"

        # Test str() conversion
        assert str(amr) == "[Adaptive mesh refinement] applied algorithm"

    def test_setting_algorithm_type_exc(self):
        """
        Tests that an AppliedAlgorithm instantiated without an Algorithm or with an invalid Algorithm instance raises
        an exception
        """
        # Tests applied algorithm defined without algorithm
        with pytest.raises(AttributeError, match="AppliedAlgorithm 'algorithm' attribute is not defined \\(mandatory\\)."):
            app_algo_exc = AppliedAlgorithm()  # No algorithm defined

        with pytest.raises(AttributeError, match="AppliedAlgorithm 'algorithm' attribute is not a valid Algorithm "
                                                 "object."):
            app_algo_exc = AppliedAlgorithm(algorithm=[9, 4.5, 2, 0.2])  # Invalid algorithm

    def test_setting_applied_algorithm_impl_details(self):
        """
        Tests setting applied algorithm implementation details property
        """
        algo = Algorithm(algo_type=AlgoType.Godunov)
        app_algo = AppliedAlgorithm(algorithm=algo)

        # ---------------- Tests invalid applied algorithm implementation details setting exception ------------------ #
        with pytest.raises(AttributeError, match="AppliedAlgorithm 'implementation_details' property is not a valid string."):
            app_algo.implementation_details = 0
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        app_algo.implementation_details = "Description of the best implementation of the Godunov scheme on Earth"
        assert app_algo.implementation_details == "Description of the best implementation of the Godunov scheme on Earth"

    def test_appalgo_equality(self):
        """
        Tests rich comparison method AppliedAlgorithm.__eq__()
        """
        amr_algo = Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement)
        amr = AppliedAlgorithm(algorithm=amr_algo, details="Implementation #3")

        # Different UUID => not equals
        assert amr != AppliedAlgorithm(algorithm=amr_algo, details=amr.implementation_details)

        # Different algorithm => not equals
        algo2 = Algorithm(algo_type=AlgoType.ParticleMesh)
        assert amr != AppliedAlgorithm(algorithm=algo2, details=amr.implementation_details, uid=amr.uid)

        # Different implementation details => not equals
        assert amr != AppliedAlgorithm(algorithm=amr_algo, uid=amr.uid, details="Alt. details")

        # Identical applied algorithm
        assert amr == AppliedAlgorithm(algorithm=amr_algo, uid=amr.uid, details=amr.implementation_details)


class TestParameterSetting(object):
    def test_psetting_init(self):
        """
        Tests ParameterSetting instance initialisation
        """
        # ParameterSetting code initialisation
        lmin = InputParameter(key="levelmin", name="lmin")
        psetting = ParameterSetting(input_param=lmin, value=7, visibility="not_displayed")

        assert psetting.input_parameter == lmin
        assert psetting.visibility == ParameterVisibility.NOT_DISPLAYED
        assert psetting.unit == U.none

        # Test str() conversion
        assert str(psetting) == "[lmin = 7] parameter setting"

    def test_setting_psetting_inpparam_exc(self):
        """
        Tests that a ParameterSetting instantiated without an InputParameter or an invalid InputParameter object raises
        an exception
        """
        # Tests parameter setting defined without input parameter
        with pytest.raises(AttributeError, match="ParameterSetting 'input_param' attribute is not defined "
                                                 "\\(mandatory\\)."):
            ps_exc = ParameterSetting()  # No input parameter defined

        # Tests parameter setting 'input_param' attribute initialisation exception
        with pytest.raises(AttributeError, match="Parameter setting 'input_param' attribute is not a valid "
                                                 "InputParameter object."):
            ps_exc = ParameterSetting(input_param=("Apple", 3, True))  # Invalid input parameter

    def test_psetting_value_exc(self):
        """
        Tests setting parameter setting 'value' property
        """
        ip = InputParameter(name="rho_saddle")
        # ---------------------- Tests parameter setting instantiation without value exception ----------------------- #
        with pytest.raises(AttributeError, match="ParameterSetting 'value' attribute is not defined \\(mandatory\\)."):
            psetting_exc = ParameterSetting(input_param=ip)
        # ------------------------------------------------------------------------------------------------------------ #

        # --------------------------- Tests invalid parameter setting value setting exception ------------------------ #
        invalid_psetting_value_err = "ParameterSetting 'value' attribute is not a valid string / integer / float " \
                                     "/ boolean value"
        with pytest.raises(AttributeError, match=invalid_psetting_value_err) as e:
            psetting_exc = ParameterSetting(input_param=ip, value=("Star", "Planet", "Galaxy"))

        with pytest.raises(AttributeError, match=invalid_psetting_value_err):
            psetting = ParameterSetting(input_param=ip, value=0.5)
            psetting_exc = psetting.value = [-1, "Black hole", False]
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        psetting = ParameterSetting(input_param=ip, value=0.5)
        assert type(psetting.value) is float and psetting.value == 0.5 and psetting.value_type == DataType.REAL
        for true_val in [True, "true", "True", "TRUE", ".true."]:
            psetting.value = true_val
            assert type(psetting.value) is bool and psetting.value is True and psetting.value_type == DataType.BOOLEAN
        for false_val in [False, "false", "False", "FALSE", ".false."]:
            psetting.value = false_val
            assert type(psetting.value) is bool and psetting.value is False and psetting.value_type == DataType.BOOLEAN

        psetting.value = "banana"
        assert type(psetting.value) is str and psetting.value == "banana" and psetting.value_type == DataType.STRING
        psetting.value = 4.256
        assert type(psetting.value) is float and psetting.value == 4.256 and psetting.value_type == DataType.REAL
        psetting.value = 58.0
        assert type(psetting.value) is int and psetting.value == 58 and psetting.value_type == DataType.INTEGER
        psetting.value = "12.0"
        assert type(psetting.value) is int and psetting.value == 12 and psetting.value_type == DataType.INTEGER
        psetting.value = "3.584e2"
        assert type(psetting.value) is float and psetting.value == 358.4 and psetting.value_type == DataType.REAL
        psetting.value = "-254"
        assert type(psetting.value) is int and psetting.value == -254 and psetting.value_type == DataType.INTEGER

    def test_setting_psetting_unit(self):
        """
        Tests setting parameter setting unit property
        """
        ip = InputParameter(name="rho_saddle")
        psetting = ParameterSetting(input_param=ip, value=1.0e-6, unit=U.H_cc)

        # --------------------------- Tests invalid parameter setting unit setting exception ------------------------- #
        with pytest.raises(AttributeError, match="Parameter setting 'unit' property error : Unknown unit name "
                                                 "'fairies'."):
            psetting.unit = "fairies"

        invalid_psetting_unit_err = "Parameter setting 'unit' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=invalid_psetting_unit_err):
            psetting.unit = ""

        with pytest.raises(AttributeError, match=invalid_psetting_unit_err):
            psetting.unit = ["kg", "V", "cm", "Gauss"]
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        psetting.unit = U.g_cc
        assert psetting.unit == U.g_cc

    def test_setting_psetting_visibility(self):
        """
        Tests setting parameter setting visibility property
        """
        ip = InputParameter(name="rho_saddle")
        psetting = ParameterSetting(input_param=ip, value=1.0e-6, visibility=ParameterVisibility.NOT_DISPLAYED)

        # ----------------------- Tests invalid parameter setting visibility setting exception ----------------------- #
        exc_msg = "Parameter setting 'visibility' property error : No ParameterVisibility defined with the key " \
                  "'custom_display'."
        with pytest.raises(AttributeError, match=exc_msg):
            psetting.visibility = "custom_display"

        with pytest.raises(AttributeError, match="ParameterSetting 'visibility' attribute is not a valid "
                                                 "ParameterVisibility enum value."):
            psetting.visibility = {"a": 0.45}
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        psetting.visibility = ParameterVisibility.ADVANCED_DISPLAY.key
        assert psetting.visibility == ParameterVisibility.ADVANCED_DISPLAY

    def test_param_setting_equality(self):
        """
        Tests rich comparison method ParameterSetting.__eq__()
        """
        dx_min = InputParameter(key="dx_min", name="min_resolution")
        psetting = ParameterSetting(input_param=dx_min, value=12.5, unit=U.kpc,
                                    visibility=ParameterVisibility.ADVANCED_DISPLAY)

        # Different UUID => not equals
        assert psetting != ParameterSetting(input_param=dx_min, value=psetting.value, unit=psetting.unit,
                                            visibility=psetting.visibility)

        # Different input parameter => not equals
        lmin = InputParameter(key="lmin", name="levelmin")
        assert psetting != ParameterSetting(input_param=lmin, value=psetting.value, unit=psetting.unit,
                                            visibility=psetting.visibility, uid=psetting.uid)

        # Different value => not equals
        assert psetting != ParameterSetting(input_param=dx_min, value=13.3, unit=psetting.unit,
                                            visibility=psetting.visibility, uid=psetting.uid)

        # Different unit => not equals
        assert psetting != ParameterSetting(input_param=dx_min, value=psetting.value, unit=U.km,
                                            visibility=psetting.visibility, uid=psetting.uid)

        # Different visibility => not equals
        assert psetting != ParameterSetting(input_param=dx_min, value=psetting.value, unit=psetting.unit,
                                            visibility=ParameterVisibility.NOT_DISPLAYED, uid=psetting.uid)

        # Identical parameter setting
        assert psetting == ParameterSetting(input_param=dx_min, value=psetting.value, unit=psetting.unit,
                                            visibility=psetting.visibility, uid=psetting.uid)

    def test_param_setting_galactica_validity_checks(self, caplog):
        """
        Tests parameter setting Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        oredring_inpp = InputParameter(key="decomp", name="Ordering")
        psetting = ParameterSetting(input_param=oredring_inpp, value="hilbert",
                                    visibility=ParameterVisibility.ADVANCED_DISPLAY)

        # Parameter setting string value too long
        psetting.value = "This is a value way too long for a simple 'ordering' RAMSES input parameter"
        psetting.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ordering = This is a value way too long for a simple 'ordering' RAMSES "
                                            "input parameter] parameter setting Galactica string value is too long "
                                            "(max. 64 characters).")


class TestResolvedPhysicalProcess(object):
    def test_res_physproc_init(self):
        """
        Tests ResolvedPhysicalProcess instance initialisation
        """
        # resolved physical process initialisation
        phys = PhysicalProcess(physics=Physics.AGNFeedback)
        res_agn_fb = ResolvedPhysicalProcess(physics=phys, details="Implemnt. #10")

        assert res_agn_fb.implementation_details == "Implemnt. #10"

        # Test str() conversion
        assert str(res_agn_fb) == "[AGN feedback] resolved physical process"

    def test_setting_physics_exc(self):
        """
        Tests that an ResolvedPhysicalProcess instantiated without a PhysicalProcess or with an invalid PhysicalProcess
        instance raises an exception
        """
        # Tests resolved physical process defined without physical process
        with pytest.raises(AttributeError, match="ResolvedPhysicalProcess 'physics' attribute is not defined "
                                                 "\\(mandatory\\)."):
            res_phys_exc = ResolvedPhysicalProcess()  # No physical process defined

        with pytest.raises(AttributeError, match="ResolvedPhysicalProcess 'physics' attribute is not a valid "
                                                 "PhysicalProcess object."):
            res_phys_exc = ResolvedPhysicalProcess(physics={"B": ["my", "physics"]})  # Invalid physical process

    def test_setting_resolved_physics_impl_details(self):
        """
        Tests setting ResolvedPhysicalProcess implementation details property
        """
        phys = PhysicalProcess(physics=Physics.AGNFeedback)
        res_agn_fb = ResolvedPhysicalProcess(physics=phys)

        # ---------------- Tests invalid applied algorithm implementation details setting exception ------------------ #
        with pytest.raises(AttributeError, match="ResolvedPhysicalProcess 'implementation_details' property is not a "
                                                 "valid string."):
            res_agn_fb.implementation_details = {"are": "you crazy ?"}
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        res_agn_fb.implementation_details = "Implementation following Dubois & Teyssier 2008"
        assert res_agn_fb.implementation_details == "Implementation following Dubois & Teyssier 2008"

    def test_resolved_physics_equality(self):
        """
        Tests rich comparison method ResolvedPhysicalProcess.__eq__()
        """
        agn = PhysicalProcess(physics=Physics.AGNFeedback)
        res_agn_fb = ResolvedPhysicalProcess(physics=agn, details="Implemnt. #10")

        # Different UUID => not equals
        assert res_agn_fb != ResolvedPhysicalProcess(physics=agn, details=res_agn_fb.implementation_details)

        # Different physical process => not equals
        sf = PhysicalProcess(physics=Physics.StarFormation)
        assert res_agn_fb != ResolvedPhysicalProcess(physics=sf, details=res_agn_fb.implementation_details,
                                                     uid=res_agn_fb.uid)

        # Different implementation details => not equals
        assert res_agn_fb != ResolvedPhysicalProcess(physics=res_agn_fb.physical_process, details="Alt. det",
                                                     uid=res_agn_fb.uid)

        # IDentical resolved physical process
        assert res_agn_fb == ResolvedPhysicalProcess(physics=agn, details=res_agn_fb.implementation_details,
                                                     uid=res_agn_fb.uid)


class TestExperiment(object):
    def test_experiment_init(self):
        """
        Tests Experiment instance initialisation
        """
        # Simulation initialisation
        cfg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "config.yml")
        exp = Experiment(name="My experiment énorme", alias="EXP_1", directory_path="/path/to/my/experiment",
                         description="""This is a pretty long description for my experiment""",
                         config_file=YamlFile.load_file(cfg_filepath))

        assert isinstance(exp.name, str)
        assert exp.alias == "EXP_1"
        assert isinstance(exp.alias, str)
        assert exp.name == "My experiment énorme"
        assert exp.description == "This is a pretty long description for my experiment"
        assert exp.directory_path == "/path/to/my/experiment"
        assert isinstance(exp.configuration_file, YamlFile) and \
               exp.configuration_file.file_md5sum == "ea17280e7d8cdf602fa7e5a436f00733"

    def test_setting_exp_name_exc(self):
        """
        Tests that an experiment defined without name raises an exception
        """
        # Tests experiment defined without name
        with pytest.raises(AttributeError, match="Experiment 'name' attribute is not defined \\(mandatory\\)."):
            exp_exc = Experiment()  # No experiment name defined

        exp = Experiment(name="My Experiment")

        # ------------- Tests invalid experiment name setting exception ---------------------- #
        empty_name_err = "Experiment 'name' property is not a valid \\(non-empty\\) string"
        with pytest.raises(AttributeError, match=empty_name_err):
            exp.name = -1

        with pytest.raises(AttributeError, match=empty_name_err):
            exp.name = ""
        # ------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        exp.name = "The best experiment ever"

    def test_setting_experiment_alias_exc(self, caplog):
        """
        Tests setting an experiment alias

        caplog: captured log PyTest fixture
        """
        exp = Experiment(name="My super experiment")

        # Tests invalid experiment alias setting exception
        with pytest.raises(AttributeError, match="Experiment 'alias' property is not a valid string"):
            simu_exc = Experiment(name="My super experiment", alias=("Dog", 3))

        with pytest.raises(AttributeError, match="Experiment 'alias' property is not a valid string"):
            exp.alias = -2.45

        # Valid => should not raise any exception
        exp.alias = "TOP_EXP_1"

        # Valid but should log a warning
        exp.alias = "ULTRA_AWESOMENESS"
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "Experiment 'alias' attribute is too long (max 16 characters).")

    def test_setting_experiment_description(self):
        """
        Tests setting experiment description property
        """
        exp = Experiment(name="My super experiment")

        # ------------------------ Tests invalid experiment description setting exception ---------------------------- #
        with pytest.raises(AttributeError, match="Experiment 'description' property is not a valid string"):
            exp.description = 0

        # Valid => should not raise any exception
        exp.description = "My experiment description"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_experiment_dirpath(self):
        """
        Tests setting experiment directory path property
        """
        exp = Experiment(name="My super simu")

        # ----------------------- Tests invalid experiment directory path setting exception -------------------------- #
        with pytest.raises(AttributeError, match="Experiment 'directory_path' property is not a valid string"):
            exp.directory_path = (-1, ["Banana", "Apple"])

        # Valid => should not raise any exception
        exp.directory_path = "/better/path/to/my/experiment"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_experiment_config_file(self):
        """
        Tests setting experiment configuration file property
        """
        exp = Experiment(name="My super simu")

        # ---------------------- Tests invalid experiment configuration file setting exception ----------------------- #
        with pytest.raises(AttributeError, match="Experiment 'configuration_file' property is neither None nor a valid "
                                                 "JsonFile, YamlFile or AsciiFile instance."):
            exp.configuration_file = {"A": -23.256}

        # Valid => should not raise any exception
        exp.configuration_file = None
        cfg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "test_implode_2D.ini")
        exp.configuration_file = AsciiFile.load_file(cfg_filepath)
        # ------------------------------------------------------------------------------------------------------------ #

    def test_experiment_equality(self):
        """
        Tests rich comparison method Experiment.__eq____()
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        cfg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "config.yml")
        exp = Experiment(name="My experiment énorme", alias="EXP_1", description="experiment description",
                         directory_path="/path/to/my/experiment", config_file=YamlFile.load_file(cfg_filepath))
        exp._protocol = ramses

        # Different UUID => not equals
        assert exp != Experiment(name=exp.name, alias=exp.alias,
                                 description=exp.description, directory_path=exp.directory_path,
                                 config_file=YamlFile.load_file(cfg_filepath))

        # Different alias => not equals
        assert exp != Experiment(name=exp.name, alias="MY_EXP_Q", description=exp.description, uid=exp.uid,
                                 directory_path=exp.directory_path, config_file=YamlFile.load_file(cfg_filepath))

        # Different name => not equals
        assert exp != Experiment(name="other name", alias=exp.alias, description=exp.description, uid=exp.uid,
                                 directory_path=exp.directory_path, config_file=YamlFile.load_file(cfg_filepath))

        # Different description => not equals
        assert exp != Experiment(name=exp.name, alias=exp.alias, description="Alt. description", uid=exp.uid,
                                 directory_path=exp.directory_path, config_file=YamlFile.load_file(cfg_filepath))

        # Different directory path => not equals
        assert exp != Experiment(name=exp.name, alias=exp.alias, description=exp.description, uid=exp.uid,
                                 directory_path="/alt/path/to/experiment", config_file=YamlFile.load_file(cfg_filepath))

        # Identical experiments
        e = Experiment(name=exp.name, alias=exp.alias, description=exp.description, uid=exp.uid,
                       directory_path=exp.directory_path, config_file=YamlFile.load_file(cfg_filepath))
        e._protocol = ramses
        assert exp == e

        # Add applied algorithms
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, description="AMR descr"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.Godunov, description="Godunov scheme"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.HLLCRiemann, description="HLLC Riemann solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.PoissonMultigrid, description="Multigrid Poisson solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.ParticleMesh, description="PM solver"))
        aa1 = exp.applied_algorithms.add(AppliedAlgorithm(algorithm=ramses.algorithms["Adaptive mesh refinement"],
                                                          details="My implementation"))
        aa2 = exp.applied_algorithms.add(AppliedAlgorithm(algorithm=ramses.algorithms["Harten-Lax-van Leer-Contact Riemann solver"],
                                                          details="My second implementation"))
        # Applied algorithms list differ => not equals
        e.applied_algorithms.add(aa1)
        assert exp != e
        # Identical experiments
        e.applied_algorithms.add(aa2)
        assert exp == e

        # Add parameter settings
        ramses.input_parameters.add(InputParameter(key="levelmin", name="Lmin",
                                                   description="min. level of AMR refinement"))
        ramses.input_parameters.add(InputParameter(key="levelmax", name="Lmax",
                                                   description="max. level of AMR refinement"))
        ps1 = exp.parameter_settings.add(ParameterSetting(input_param=ramses.input_parameters["Lmin"], value=8,
                                                          visibility=ParameterVisibility.BASIC_DISPLAY))
        ps2 = exp.parameter_settings.add(ParameterSetting(input_param=ramses.input_parameters["Lmax"], value=12,
                                                          visibility=ParameterVisibility.BASIC_DISPLAY))
        # Parameter settings list differ => not equals
        e.parameter_settings.add(ps1)
        assert exp != e
        # Identical experiments
        e.parameter_settings.add(ps2)
        assert exp == e

        # Add snapshots
        sn = exp.snapshots.add(Snapshot(name="First pericenter", description="Snapshot of the first pericenter",
                                        time=(100.25, U.Myr), physical_size=(1.0, U.pc),
                                        directory_path="/path/to/my/snapshot"))
        # Snapshot list differ => not equals
        assert e != exp
        # Identical experiments
        e.snapshots.add(sn)
        assert e == exp

        # Add generic results
        gres = exp.generic_results.add((GenericResult(name="Experiment main result")))
        # Generic result list differ => not equals
        assert exp != e
        # Identical experiments
        e.generic_results.add(gres)
        assert exp == e

    def test_experiment_galactica_validity_checks(self, caplog):
        """
        Tests experiment Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        ramses = SimulationCode(name="Ramses 3.0 (Hydro)", code_name="RAMSES", alias="RAMSES_CODE")
        exp = Simulation(ramses, name="My simu (hydro)", execution_time="2020-06-01 20:00:00")

        # No alias defined
        exp.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My simu (hydro)' simulation Galactica alias is missing.")

        # Alias too long
        exp.alias = "MY_ALIAS_123456789_10_fgkjhfgoiurt_jnsdfnsdfdgdfgdfgdfgdfgkfgdfgdfgdgf"
        exp.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My simu (hydro)' simulation Galactica alias is too long (max. 64 "
                                            "characters).")

        # Invalid Galactica alias
        exp.alias = "MY_DUMMY_"
        exp.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My simu (hydro)' simulation Galactica alias is not valid (The alias can "
                                            "contain capital letters, digits and '_' only. It must start with a "
                                            "capital letter and cannot end with a '_'.)")

        # Result duplicates (same result names)
        exp.alias = "MY_ALIAS_123456789_10"  # => Ok
        exp.generic_results.add(GenericResult(name="My_result"))
        exp.snapshots.add(Snapshot(name="My_result"))
        exp.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My_result' generic result and 'My_result' snapshot results share the "
                                            "same name. They must be unique.")


class TestSimulation(object):
    def test_simu_init(self):
        """
        Tests Simulation instance initialisation
        """
        # Simulation initialisation
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(name="My simulation énorme", alias="SIMU_1", simu_code=ramses,
                          description="""This is a pretty long description for my simulation""",
                          directory_path="/path/to/my/simulation", execution_time="2020-03-01 18:45:30")

        assert isinstance(ramses.name, str)
        assert isinstance(simu.name, str)
        assert simu.alias == "SIMU_1"
        assert isinstance(simu.alias, str)
        assert simu.name == "My simulation énorme"
        assert simu.description == "This is a pretty long description for my simulation"
        assert simu.directory_path == "/path/to/my/simulation"
        assert isinstance(simu.execution_time, str)
        assert simu.execution_time == "2020-03-01 18:45:30"

        # Test str() conversion
        assert str(simu) == "'My simulation énorme' simulation"

    def test_simulation_simu_code_setting_exc(self):
        """
        Tests setting a simulation code into a simulation object
        """
        # Tests simulation defined without code
        with pytest.raises(AttributeError, match="Undefined simulation code for 'My super simulation' Simulation."):
            simu_exc = Simulation(name="My super simulation")  # No simulation code defined

        # Tests invalid simulation_code initialisation exception
        invalid_simu_code_err = "Simulation 'simulation_code' attribute is not a valid SimulationCode instance."
        with pytest.raises(AttributeError, match=invalid_simu_code_err):
            simu_exc = Simulation(name="My simu", simu_code=-5)  # Invalid simulation code
        with pytest.raises(AttributeError, match=invalid_simu_code_err):
            simu_exc = Simulation(-5, name="My simu")  # Invalid simulation code

        # Simulation initialisation : valid => should not raise any exception
        ramses = SimulationCode(name="Ramses", code_name="Ramses")
        simulation = Simulation(name="My simu", simu_code=ramses)
        simulation2 = Simulation(ramses, name="My simu")

    def test_setting_simulation_execution_time(self):
        """
        Tests setting simulation execution time property
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu")

        # ------------------------ Tests invalid experiment description setting exception ---------------------------- #
        with pytest.raises(AttributeError, match="Simulation 'execution_time' property is not a valid datetime string."):
            simu.execution_time = {"a": 1.435}

        # Valid => should not raise any exception
        simu.execution_time = "2020-04-26 10:52:46"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_simu_equality(self):
        """
        Tests rich comparison method Simulation.__eq____()
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme", execution_time="2020-03-01 18:45:30")

        # Different simulation code => not equals
        gadget = SimulationCode(name="Gadget 4.2.1", code_name="Gadget")
        assert simu != Simulation(name=simu.name, simu_code=gadget, uid=simu.uid, execution_time=simu.execution_time)

        # Different execution time => not equals
        assert simu != Simulation(name=simu.name, simu_code=simu.simulation_code, uid=simu.uid,
                                  execution_time="2018-09-14 15:48:32")

        # Identical simulations
        s = Simulation(name=simu.name, simu_code=simu.simulation_code, uid=simu.uid, execution_time=simu.execution_time)
        assert simu == s

        # Add resolved physical processes
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.StarFormation, description="descr sf"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.Hydrodynamics, description="descr hydro"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SelfGravity, description="descr self G"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SupernovaeFeedback, description="SN feedback"))
        rpp1 = simu.resolved_physics.add(ResolvedPhysicalProcess(physics=ramses.physical_processes["Star formation"],
                                                                 details="Star formation specific implementation"))
        rpp2 = simu.resolved_physics.add(ResolvedPhysicalProcess(physics=ramses.physical_processes["Self-gravity"],
                                                                 details="self-gravity specific implementation"))
        # Resolved physical processes list differ => not equals
        s.resolved_physics.add(rpp1)
        assert simu != s
        # Identical simulations
        s.resolved_physics.add(rpp2)
        assert simu == s

        # Add post-processing runs
        hop = PostProcessingCode(name="Hop", code_name="HOP")
        pprun = simu.post_processing_runs.add(PostProcessingRun(name="My post-processing run", ppcode=hop))
        # Post-processing runs differ => not equals
        assert simu != s
        # Identical simulations
        s.post_processing_runs.add(pprun)
        assert s == simu
        # ------------------------------------------------------------------------------------------------------------ #

    def test_simulation_galactica_validity_checks(self, caplog):
        """
        Tests simulation Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        ramses = SimulationCode(name="Ramses 3.0 (Hydro)", code_name="RAMSES", alias="RAMSES_CODE")
        simu = Simulation(ramses, name="My simu (hydro)", alias="MY_SIMU")

        # No execution time defined
        simu.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My simu (hydro)' simulation Galactica execution time is not defined.")

    def test_simu_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Simulation from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu_name = "My most interesting simulation"
        cfg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "test_implode_2D.ini")
        simu = Simulation(simu_code=ramses, name=simu_name, alias="SIMU_1", description="simu description",
                          execution_time="2020-03-01 18:45:30", config_file=AsciiFile.load_file(cfg_filepath))

        # Add algortihms
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, description="AMR descr"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.Godunov, description="Godunov scheme"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.HLLCRiemann, description="HLLC Riemann solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.PoissonMultigrid, description="Multigrid Poisson solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.ParticleMesh, description="PM solver"))
        simu.applied_algorithms.add(AppliedAlgorithm(algorithm=ramses.algorithms["Adaptive mesh refinement"],
                                                     details="My implementation"))
        simu.applied_algorithms.add(AppliedAlgorithm(algorithm=ramses.algorithms["Harten-Lax-van Leer-Contact Riemann solver"],
                                                     details="My second implementation"))

        # Add input parameters
        ramses.input_parameters.add(InputParameter(key="levelmin", name="Lmin",
                                                   description="min. level of AMR refinement"))
        ramses.input_parameters.add(InputParameter(key="levelmax", name="Lmax",
                                                   description="max. level of AMR refinement"))
        boxlen_key = "boxlen"
        ramses.input_parameters.add(InputParameter(key=boxlen_key, name="Lbox",
                                                   description="Simulation domain box size"))
        simu.parameter_settings.add(ParameterSetting(input_param=ramses.input_parameters["Lmin"], value=8,
                                                     visibility=ParameterVisibility.BASIC_DISPLAY))
        simu.parameter_settings.add(ParameterSetting(input_param=ramses.input_parameters["Lmax"], value=12,
                                                     visibility=ParameterVisibility.BASIC_DISPLAY))
        simu.parameter_settings.add(ParameterSetting(input_param=ramses.input_parameters["Lbox"], value=500.0,
                                                     unit=U.kpc, visibility=ParameterVisibility.BASIC_DISPLAY))

        # Add physical processes
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.StarFormation, description="descr sf"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.Hydrodynamics, description="descr hydro"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SelfGravity, description="descr self G"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SupernovaeFeedback, description="SN feedback"))
        simu.resolved_physics.add(ResolvedPhysicalProcess(physics=ramses.physical_processes["Star formation"],
                                                          details="Star formation specific implementation"))
        simu.resolved_physics.add(ResolvedPhysicalProcess(physics=ramses.physical_processes["Self-gravity"],
                                                          details="self-gravity specific implementation"))

        # Add snapshot
        sn = Snapshot(name="First pericenter", description="Snapshot of the first pericenter",
                      time=(100.25, U.Myr), physical_size=1.0, directory_path="/path/to/my/snapshot")
        simu.snapshots.add(sn)

        # Add generic result
        simu.generic_results.add(GenericResult(name="Simu key result !"))

        # Add post-processing run
        hop = PostProcessingCode(name="Hop", code_name="HOP")
        simu.post_processing_runs.add(PostProcessingRun(name="My post-processing run", ppcode=hop))

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project épatant")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        proj_loaded = study_loaded.project
        simu_loaded = proj_loaded.simulations[simu_name]
        assert simu_loaded == simu

        # Test modifying unit to U.none, saving and reloading it => check that HDF5 file do not contain unit anymore
        loaded_lbox_psetting = simu_loaded.parameter_settings[boxlen_key]
        loaded_lbox_psetting.unit = U.none  # => Setting the unit to U.none here !
        fname2 = str(tmp_path / "study2.h5")
        study_loaded.save_HDF5(fname2)
        study_reloaded = SimulationStudy.load_HDF5(fname2)
        project_reloaded = study_reloaded.project
        simu_reloaded = project_reloaded.simulations[simu_name]
        reloaded_lbox_psetting = simu_reloaded.parameter_settings[boxlen_key]
        assert reloaded_lbox_psetting.unit == U.none


class TestPostProcessingRun(object):
    def test_simu_init(self):
        """
        Tests PostProcessingRun instance initialisation
        """
        # PostProcessingRun initialisation
        hop = PostProcessingCode(name="Hop", code_name="HOP")
        pprun = PostProcessingRun(name="My post-processing run", alias="PPRUN_1", ppcode=hop,
                                  description="This is a pretty long description for my post-processing run",
                                  directory_path="/path/to/my/simulation")

        assert pprun.alias == "PPRUN_1"
        assert pprun.name == "My post-processing run"
        assert pprun.description == "This is a pretty long description for my post-processing run"
        assert pprun.directory_path == "/path/to/my/simulation"
        assert pprun.configuration_file is None

        # Test str() conversion
        assert str(pprun) == "'My post-processing run' post-processing run"

    def test_post_processing_run_ppcode_setting_exc(self):
        """
        Tests setting a post-processing code into a post-processing run object
        """
        # Tests post-processing run defined without code
        with pytest.raises(AttributeError, match="Undefined post-processing code for 'My super post-processing run' "
                                                 "Post-processing run."):
            pprun_exc = PostProcessingRun(name="My super post-processing run")  # No post-processing code defined

        # Tests invalid post-processing run_code initialisation exception
        invalid_ppcode_err = "PostProcessingRun 'postpro_code' attribute is not a valid PostProcessingCode instance."
        with pytest.raises(AttributeError, match=invalid_ppcode_err):
            pprun_exc = PostProcessingRun(name="My post-processing run", ppcode=-5)  # Invalid post-processing run code
            pprun_exc2 = PostProcessingRun(-5, name="My post-processing run")  # Invalid post-processing run code

        # PostProcessingRun initialisation : valid => should not raise any exception
        RADMC = PostProcessingCode(name="RadMC", code_name="RADMC")
        pprun = PostProcessingRun(name="My run", ppcode=RADMC)
        pprun2 = PostProcessingRun(RADMC, name="My run")

    def test_pprun_equality(self):
        """
        Tests rich comparison method PostProcessingRun.__eq____()
        """
        hop = PostProcessingCode(name="Hop", code_name="HOP")
        pprun = PostProcessingRun(hop, name="My run éblouissant")

        # Different post-processing code => not equals
        adaptaHop = PostProcessingCode(name="AdaptaHOP v2.1", code_name="AdaptaHop")
        assert pprun != PostProcessingRun(name=pprun.name, ppcode=adaptaHop, uid=pprun.uid)

        # Identical post-processing runs
        ppr = PostProcessingRun(name=pprun.name, ppcode=pprun.postpro_code, uid=pprun.uid)
        assert pprun == ppr

    def test_pprun_hdf5_io(self, tmp_path):
        """
        Tests saving/loading PostProcessingRun from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path PyTest fixture
        """
        hop = PostProcessingCode(name="Hop", code_name="HOP")
        cfg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "test_implode_2D.ini")
        pprun = PostProcessingRun(hop, name="My run éblouissant", alias="PPRUN_1", description="run description",
                                  config_file=AsciiFile.load_file(cfg_filepath))

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.post_processing_runs.add(pprun)
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations["My simu énorme"]
        pprun_loaded = simu_loaded.post_processing_runs["My run éblouissant"]
        assert pprun_loaded == pprun


__all__ = ["TestSimulation", "TestPostProcessingRun"]




# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import absolute_import
from numpy import pi, tan, radians, array
from .unit import Unit, UnitError


# --------------------------------------------------------------------------------------------------------------------#
# Dimensionless units
none = Unit.create_unit('none',       dims=(0, 0, 0, 0, 0, 0, 0, 0), descr="Unscaled dimensionless unit")
percent = Unit.create_unit('percent', base_unit=0.01 * none, descr="One hundredth of unity", latex="\\%")


# --------------------------------------------------------------------------------------------------------------------#
# Time units
s = Unit.create_unit('s',             dims=(0, 0, 1, 0, 0, 0, 0, 0), descr="Second : base time unit")
min = Unit.create_unit('min',         base_unit=60. * s,         descr="Minute", latex="\\textrm{min}")
hour = Unit.create_unit('hour',       base_unit=3600. * s,       descr="Hour", latex="\\textrm{h}")
day = Unit.create_unit('day',         base_unit=24. * hour,      descr="Day", latex="\\textrm{day}")
sid_day = Unit.create_unit('sid_day', base_unit=86164.09053 * s, descr="Sidereal day : Earth full rotation time",
                                                                 latex="\\textrm{T}_{\\textrm{sid}}")
year = Unit.create_unit('year',       base_unit=365.25 * day,    descr="Year", latex="\\textrm{yr}")
kyr = Unit.create_unit('kyr',         base_unit=1.0E3 * year,    descr="kyr : millenium", latex="\\textrm{kyr}")
Myr = Unit.create_unit('Myr',         base_unit=1.0E6 * year,    descr="Megayear : million years", latex="\\textrm{Myr}")
Gyr = Unit.create_unit('Gyr',         base_unit=1.0E9 * year,    descr="Gigayear : trillion years", latex="\\textrm{Gyr}")


# --------------------------------------------------------------------------------------------------------------------#
# Distance units
m = Unit.create_unit('m',               dims=(0, 1, 0, 0, 0, 0, 0, 0), descr="Meter : base length unit")
um = Unit.create_unit('um',             base_unit=1.0E-6 * m,         descr="Micron", latex="\\mu m")
mm = Unit.create_unit('mm',             base_unit=1.0E-3 * m,         descr="Millimeter", latex="\\textrm{mm}")
cm = Unit.create_unit('cm',             base_unit=1.0E-2 * m,         descr="Centimeter", latex="\\textrm{cm}")
nm = Unit.create_unit('nm',             base_unit=1.0E-9 * m,         descr="Nanometer", latex="\\textrm{nm}")
km = Unit.create_unit('km',             base_unit=1000.0 * m,         descr="Kilometer", latex="\\textrm{km}")
Angstrom = Unit.create_unit('Angstrom', base_unit=1.0E-10 * m,        descr="Angstrom: 10**-10 m", latex="\\AA")
au = Unit.create_unit('au',             base_unit=1.495978707e8 * km, descr="Astronomical unit", latex="\\textrm{au}")
pc = Unit.create_unit('pc',             base_unit=au / tan(radians(1. / 3600.)), descr="Parsec", latex="\\textrm{pc}")  # 3.085677e16 * m
kpc = Unit.create_unit('kpc',           base_unit=1.0E3 * pc,         descr="Kiloparsec", latex="\\textrm{kpc}")
Mpc = Unit.create_unit('Mpc',           base_unit=1.0E6 * pc,         descr="Megaparsec", latex="\\textrm{Mpc}")
Gpc = Unit.create_unit('Gpc',           base_unit=1.0E9 * pc,         descr="Gigaparsec", latex="\\textrm{Gpc}")
Rsun = Unit.create_unit('Rsun',         base_unit=6.95508E5 * km,     descr="Solar radius", latex="\\textrm{R}_{\\odot}")


# ------------------------------------------------------------------------------------------------------------------- #
# Volume units
cm3 = Unit.create_unit('cm3',           base_unit=cm**3,              descr="Cubic centimeter", latex="\\textrm{cm}^{3}")
m3 = Unit.create_unit('m3',             base_unit=m**3,               descr="Cubic meter", latex="\\textrm{m}^{3}")
pc3 = Unit.create_unit('pc3',           base_unit=pc**3,              descr="Cubic parsec", latex="\\textrm{pc}^{3}")
kpc3 = Unit.create_unit('kpc3',         base_unit=pc**3,              descr="Cubic kiloparsec", latex="\\textrm{kpc}^{3}")


# ------------------------------------------------------------------------------------------------------------------- #
# Frequency units
Hz = Unit.create_unit('Hz', base_unit=1.0 / s, descr="Hertz : frequency unit", latex="\\textrm{Hz}")
kHz = Unit.create_unit('kHz', base_unit=1.0E3 * Hz, descr="kilo-Hertz : frequency unit", latex="\\textrm{kHz}")
MHz = Unit.create_unit('MHz', base_unit=1.0E6 * Hz, descr="mega-Hertz : frequency unit", latex="\\textrm{MHz}")
GHz = Unit.create_unit('GHz', base_unit=1.0E9 * Hz, descr="giga-Hertz : frequency unit", latex="\\textrm{GHz}")


# ------------------------------------------------------------------------------------------------------------------- #
# Surface units
barn = Unit.create_unit('barn', 1.0E-28 * m**2, descr="barn: surface unit used in HEP", latex="\\textrm{barn}")


# ------------------------------------------------------------------------------------------------------------------- #
# Velocity units
m_s = Unit.create_unit('m_s',           base_unit=m /s,        descr="Meters per second",
                                                               latex="\\textrm{m}\\cdot\\textrm{s}^{-1}")
km_s = Unit.create_unit('km_s',         base_unit=1.0e3 * m_s, descr="kilometers per second",
                                                               latex="\\textrm{km}\\cdot\\textrm{s}^{-1}")


# --------------------------------------------------------------------------------------------------------------------#
# Mass units
kg = Unit.create_unit('kg',         coeff=1.0, dims=(1, 0, 0, 0, 0, 0, 0, 0), descr="Kilogram : base mass unit", latex="\\textrm{kg}")
g = Unit.create_unit('g',           base_unit=1.0E-3 * kg,    descr="Gram")
t = Unit.create_unit('t',           base_unit=1.0E3 * kg,     descr="Metric ton")
mH = Unit.create_unit('mH',         base_unit=1.66E-24 * g,   descr="Hydrogen atomic mass", latex="m_{\\textrm{H}}")
Msun = Unit.create_unit('Msun',     base_unit=1.9889E30 * kg, descr="Solar mass", latex="\\textrm{M}_{\\odot}")
Mearth = Unit.create_unit('Mearth', base_unit=5.9722E24 * kg, descr="Earth mass", latex="\\textrm{M}_{\\oplus}")


# ------------------------------------------------------------------------------------------------------------------- #
# Angular measurements
rad = Unit.create_unit('rad', dims=(0, 0, 0, 0, 0, 0, 1, 0),
                       descr="radian: angular measurement (ratio lengh / radius of an arc)", latex="\\textrm{rad}")
degree = Unit.create_unit('deg',          base_unit=(pi / 180.0) * rad, latex="^{\\circ}",
                          descr="degree: angular measurement corresponding to 1/360 of a full rotation")
hourangle = Unit.create_unit('hourangle', base_unit=15.0 * degree,
                             descr="hour angle: angular measurement with 24 in a full circle")
arcmin = Unit.create_unit('arcmin',       base_unit=degree / 60.0, descr="arc minute: 1/60 of a hour angle", latex="'")
arcsec = Unit.create_unit('arcsec',       base_unit=degree/3600.0, descr="arc second: 1/60 of an arcminute", latex="\"")
sr = Unit.create_unit('sr',               base_unit=rad ** 2, descr="Steradian: solid angle (SI) unit")

# --------------------------------------------------------------------------------------------------------------------#
# Force units
N = Unit.create_unit('N',       base_unit=kg * m / s ** 2, descr="Newton : (SI) force unit")
dyne = Unit.create_unit('dyne', base_unit=g * cm / s ** 2, descr="dyne : (CGS) force unit", latex="\\textrm{dyne}")

# --------------------------------------------------------------------------------------------------------------------#
# Temperature unit
K = Unit.create_unit('K', dims=(0, 0, 0, 1, 0, 0, 0, 0), descr="Kelvin : base temperature unit")


# --------------------------------------------------------------------------------------------------------------------#
# Energy units
J = Unit.create_unit('J',     base_unit=kg * (m / s)**2,   descr="Joule : (SI) energy unit")
W = Unit.create_unit('W',     base_unit=J / s,             descr="Watt")
erg = Unit.create_unit('erg', base_unit=g * (cm / s) ** 2, descr="erg : (CGS) energy unit", latex="\\textrm{erg}")

# --------------------------------------------------------------------------------------------------------------------#
# Pressure unit
barye = Unit.create_unit('barye', base_unit=g / (cm * s**2), descr="Barye: (CGS) pressure unit")
Pa = Unit.create_unit('Pa',       base_unit=J * m ** -3,     descr="Pascal: (SI) pressure unit", latex="\\textrm{Pa}")
hPa = Unit.create_unit('hPa',     base_unit=1.0E2 * Pa,      descr="Hectopascal", latex="\\textrm{hPa}")
kPa = Unit.create_unit('kPa',     base_unit=1.0E3 * Pa,      descr="Kilopascal", latex="\\textrm{kPa}")
bar = Unit.create_unit('bar',     base_unit=1.0E5 * Pa,      descr="Bar", latex="\\textrm{bar}")
atm = Unit.create_unit('atm',     base_unit=1.01325 * bar,   descr="atm: atmospheric pressure (101 3525 Pa)", latex="\\textrm{atm}")


# --------------------------------------------------------------------------------------------------------------------#
# Electrical units
A = Unit.create_unit('A',     dims=(0, 0, 0, 0, 1, 0, 0, 0), descr="Ampere : electric intensity base unit")
C = Unit.create_unit('C',     base_unit=A * s,               descr="Coulomb")
e = Unit.create_unit('e',     base_unit=1.602176565e-19 * C, descr="e : elementary electric charge carried by a proton")
V = Unit.create_unit('V',     base_unit=J / C,               descr="Volt")
Ohm = Unit.create_unit('Ohm', base_unit=V * s / C,           descr="Ohm", latex="\\Omega")
S = Unit.create_unit('S',     base_unit=C / (V * s),         descr="Siemens")
F = Unit.create_unit('F',     base_unit=C / V,               descr="Farad")


# --------------------------------------------------------------------------------------------------------------------#
# Magnetical units
T = Unit.create_unit('T',           base_unit=V * s / m**2,   descr="Tesla")
Gauss = Unit.create_unit('Gauss',   base_unit=1.0E-4 * T,     descr="Gauss", latex="\\textrm{G}")
mGauss = Unit.create_unit('mGauss', base_unit=1.0E-3 * Gauss, descr="Milligauss", latex="\\textrm{mG}")
uGauss = Unit.create_unit('uGauss', base_unit=1.0E-6 * Gauss, descr="Microgauss", latex="\\mu \\textrm{G}")
Henry = Unit.create_unit('Henry',   base_unit=T * m**2 / A,   descr="Henry",      latex="\\textrm{H}")


# ------------------------------------------------------------------------------------------------------------------- #
# Amount of substance
mol = Unit.create_unit('mol', dims=(0, 0, 0, 0, 0, 1, 0, 0), descr="mole: amount of a chemical substance base unit")


# ------------------------------------------------------------------------------------------------------------------- #
# Illumination
cd = Unit.create_unit('cd',     dims=(0, 0, 0, 0, 0, 0, 0, 1), descr="Candela: base luminous intensity unit")
lm = Unit.create_unit('lm',     base_unit=cd * sr,             descr="Lumen")
lx = Unit.create_unit('lx',     base_unit=lm / m**2,           descr="Lux")
Lsun = Unit.create_unit('Lsun', base_unit=3.846E26 * W,        descr="Solar luminosity", latex="\\textrm{L}_{\\odot}")


# ------------------------------------------------------------------------------------------------------------------- #
# Spectral density
Jy = Unit.create_unit('Jy', base_unit=1.0E-26 * W / (m**2 * Hz), descr="Jansky")


# --------------------------------------------------------------------------------------------------------------------#
# Composite units
c = Unit.create_unit('c',       base_unit=2.99792458E8 * m / s,          descr="Speed of light in vacuum")
ly = Unit.create_unit('ly',     base_unit=c * year,                      descr="Light year")
eV = Unit.create_unit('eV',     base_unit=e * V,                         descr="electron-Volt")
G = Unit.create_unit('G',       base_unit=6.67428e-11 * m ** 3 / kg / (s ** 2), descr="Graviational constant")
kB = Unit.create_unit('kB',     base_unit=1.3806504E-23 * J / K,         descr="Boltzmann constant", latex="k_{\\textrm{B}}")
H = Unit.create_unit('H',       base_unit=70.0 * km / s / Mpc,           descr="Hubble's constant")
rhoc = Unit.create_unit('rhoc', base_unit=3.0 * H ** 2 / (8.0 * pi * G), descr="Friedmann's universe critical density",
                                                                         latex="\\rho_{c}")
H_cc = Unit.create_unit('H_cc', base_unit=mH / cm ** 3 / 0.76,           descr="Atoms per cubic centimeter",
                                                                         latex="\\textrm{H}/\\textrm{cc}")
g_cc = Unit.create_unit('g_cc', base_unit=g / cm ** 3,                   descr="Gram per cubic centimeter",
                                                                         latex="\\textrm{g}/\\textrm{cc}")


# --------------------------------------------------------------------------------------------------------------------#
# Planck constant
h = Unit.create_unit('h',       base_unit=6.62606957E-34 * J * s, descr="Planck Constant")
hbar = Unit.create_unit('hbar', base_unit=h / (2*pi),             descr="Reduced Planck constant", latex="\\bar{h}")

# --------------------------------------------------------------------------------------------------------------------#
# Stefan-Boltzmann constants
sigmaSB = Unit.create_unit('sigmaSB', base_unit=(2 * pi**5 * kB**4) / (15 * h**3 * c**2), descr="Stefan-Boltzmann constant",
                                                                                          latex="\\sigma_{\\textrm{SB}}")
a_R = Unit.create_unit('a_r',         base_unit=4 * sigmaSB / c, descr="Radiation constant", latex="a_{R}")


# ----- Physical type creation ----- #
Unit.create_phystype(none, 'dimensionless')
Unit.create_phystype(m, 'length')
Unit.create_phystype(m**2, 'area')
Unit.create_phystype(m3, 'volume')
Unit.create_phystype(s, 'time')
Unit.create_phystype(rad, 'angle')
Unit.create_phystype(sr, 'solid angle')
Unit.create_phystype(m_s, 'velocity')
Unit.create_phystype(m / s**2, 'acceleration')
Unit.create_phystype(Hz, 'frequency')
Unit.create_phystype(g, 'mass')
Unit.create_phystype(mol, 'amount of substance')
Unit.create_phystype(K, 'temperature')
Unit.create_phystype(N, 'force')
Unit.create_phystype(J, 'energy')  # or 'torque'
Unit.create_phystype(J/K, 'entropy')
Unit.create_phystype(J/s/m**2, 'energy flux density')
Unit.create_phystype(J/kg, 'specific energy')
Unit.create_phystype(Pa, 'pressure')  # or 'energy density'
Unit.create_phystype(W, 'power')  # or 'energy flux'
Unit.create_phystype(W/(m*K), 'thermal conductivity')
Unit.create_phystype(kg / m**3, 'volume density')
Unit.create_phystype(kg / m**2, 'surface density')
Unit.create_phystype(kg / m, 'linear density')
Unit.create_phystype(m**3 / kg, 'specific volume')
Unit.create_phystype(mol / m**3, 'molar volume')
Unit.create_phystype(kg * m / s, 'momentum/impulse')
Unit.create_phystype(kg*m**2, "moment of inertia")
Unit.create_phystype(kg * m**2 / s, 'angular momentum')
Unit.create_phystype(rad / s, 'angular velocity')
Unit.create_phystype(rad / s**2, 'angular acceleration')
Unit.create_phystype(g / (m * s), 'dynamic viscosity')
Unit.create_phystype(m ** 2 / s, 'kinematic viscosity')
Unit.create_phystype(m**-1, 'wavenumber')
Unit.create_phystype(A, 'electric current')
Unit.create_phystype(C, 'electric charge')
Unit.create_phystype(V, 'electric potential')
Unit.create_phystype(Ohm, 'electric resistance')
Unit.create_phystype(Ohm*m, 'electric resistivity')
Unit.create_phystype(S, 'electric conductance')
Unit.create_phystype(S/m, 'electric conductivity')
Unit.create_phystype(F, 'electric capacitance')
Unit.create_phystype(C * m, 'electric dipole moment')
Unit.create_phystype(A / m**2, 'electric current density')
Unit.create_phystype(V / m, 'electric field strength')
Unit.create_phystype(C / m**2, 'electric flux density')  # or 'electric displacement'
Unit.create_phystype(C / m**3, 'electric charge density')
Unit.create_phystype(F / m, 'permittivity')
Unit.create_phystype(T, 'magnetic flux density')
Unit.create_phystype(T * m**2, 'magnetic flux')
Unit.create_phystype(A / m, 'magnetic field strength')
Unit.create_phystype(Henry / m, 'magnetic permeability')
Unit.create_phystype(Henry, 'inductance')
Unit.create_phystype(cd, 'luminous intensity')
Unit.create_phystype(lm, 'luminous flux')
Unit.create_phystype(lx, 'luminous emittence')
Unit.create_phystype(W / sr, 'radiant intensity')
Unit.create_phystype(cd / m**2, 'luminance')
Unit.create_phystype(Jy, 'spectral flux density')

del pi, tan, radians


# ------------------------------------------------------------------------------------------------------------------- #
# ----------------------------------------------- Rich docstring ---------------------------------------------------- #
# ------------------------------------------------------------------------------------------------------------------- #
_mod_doc = """

Base units
^^^^^^^^^^

    +-----------+-------------------------------------------------------------------+
    |   Name    |                      Description                                  |
    +===========+===================================================================+
"""
_ns = locals()
_sk = _ns.keys()
_sk = sorted(_sk)
_su = []
for _k in _sk:
    _v = _ns[_k]
    if isinstance(_v, Unit) and _v.is_base_unit():
        _su.append(_k)

for _k in _su:
    _v = _ns[_k]
    _mod_doc += "    |%10s | %65s | \n"% (_v._name, _v._descr)
    _mod_doc += "    +-----------+-------------------------------------------------------------------+\n"

_mod_doc += """

.. seealso:: `Wikipedia : SI base unit <https://en.wikipedia.org/wiki/SI_base_unit>`_


.. _phys_constants:

Constants and common units
^^^^^^^^^^^^^^^^^^^^^^^^^^

    +-----------+---------------+-----------------------------+---------------------------------------------------""" +\
    """---------------------+
    |   Name    |      Value    | Decomposition in base units |                         Description               """ +\
    """                     |
    +===========+===============+=============================+===================================================""" +\
    """=====================+
"""

_su = []
for _k in _sk:
    _v = _ns[_k]
    if isinstance(_v, Unit) and not _v.is_base_unit():
        _su.append(_k)

for _k in _su:
    _v = _ns[_k]
    _mod_doc += "    |%10s |%14g | %27s | %70s | \n"% (_v._name, _v._coeff, _v._decompose_base_units(), _v._descr)
    _mod_doc += "    +-----------+---------------+-----------------------------+" \
                "------------------------------------------------------------------------+\n"

_rpqd = {qty: d for d, qty in Unit._phystype_unit_mapping.items()}
_pq = Unit._phystype_unit_mapping.values()
_spq = sorted(_pq)

_mod_doc += """

.. _phys_qties:

Physical quantities
^^^^^^^^^^^^^^^^^^^

    +-------------------------------+---------------------------------+
    |  Quantity                     |   Decomposition in base units   |
    +===============================+=================================+
"""
for _q in _spq:
    _d = _rpqd[_q]
    _u = Unit(dims=_d)
    _mod_doc += "    | {qty:29s} | {bu:31s} |\n".format(qty=_q, bu=_u._decompose_base_units())
    _mod_doc += "    +-------------------------------+---------------------------------+\n"

_mod_doc += "\n"

__doc__ = _mod_doc

del _mod_doc, _su, _sk, _ns
# ------------------------------------------------------------------------------------------------------------------- #


__all__ = ["Unit", "UnitError"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import uuid
import os
import shutil
import pandas as pd

from future.builtins import str
import logging

from astrophysix.utils.persistency import Hdf5StudyPersistent, Hdf5StudyFileMode
from astrophysix.utils.strings import Stringifiable
from .field import CatalogField
from .targobj import TargetObject
from ..datafiles import Datafile
from ..services import CatalogDataProcessingService
from ..utils import ObjectList, GalacticaValidityCheckMixin
from ...utils import FileUtil, FileType, DatetimeUtil, HDF5IOFile

log = logging.getLogger("astrophysix.simdm")


class Catalog(Hdf5StudyPersistent, HDF5IOFile, GalacticaValidityCheckMixin, Stringifiable):
    """
    Result object catalog class (Simulation data model)

    Parameters
    ----------
    target_object: :class:`~astrophysix.simdm.catalogs.targobj.TargetObject`
        catalog object type (mandatory)
    name: :obj:`string`
        catalog name (mandatory)
    description: :obj:`string`
        catalog description
    """
    _hsp_version = 2  # With data processing services
    def __init__(self, *args, **kwargs):
        super(Catalog, self).__init__(**kwargs)
        self._name = ""
        self._description = ""
        self._fields = ObjectList(CatalogField, "property_name")
        self._fields.add_validity_check_method(self._does_field_have_same_number_of_items)
        self._fields.add_validity_check_method(self._does_field_property_is_target_obj_prop)
        self._fields.add_deletion_handler(self._can_delete_catalog_field)
        self._datafiles = ObjectList(Datafile, "name")
        self._targobj = None
        self._proc_services = ObjectList(CatalogDataProcessingService, "hosted_service",
                                         object_addition_vcheck=(self._can_add_field_binding_in_proc_service,
                                                                 "catalog_field_bindings"))
        self._proc_services.add_validity_check_method(self._can_add_proc_service)

        # Object catalog name
        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        # ------------------------------------- Target object ------------------------------------- #
        if len(args) > 0:
            targobj = args[0]
        elif "target_object" in kwargs:
            targobj = kwargs["target_object"]
        else:
            raise AttributeError("Undefined target object for {cat!s}.".format(cat=self))

        if not isinstance(targobj, TargetObject):
            err_msg = "Catalog 'target_object' attribute is not a valid TargetObject instance."
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._targobj = targobj

        # Add deletion handler to the target object's property list
        targobj.object_properties.add_deletion_handler(self._can_delete_object_property)
        # ----------------------------------------------------------------------------------------- #

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        Catalog comparison method

        other: :class:`~astrophysix.simdm.catalogs.catalog.Catalog`
            Other catalog instance to compare to
        """
        if not super(Catalog, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        # Compare target object
        if self._targobj != other.target_object:
            return False

        #  Compare datafiles
        if self._datafiles != other.datafiles:
            return False

        if self._description != other.description:
            return False

        # Compare catalog fields
        if self._fields != other.catalog_fields:
            return False

        # Compare data processing service list
        if self._proc_services != other.processing_services:
            return False

        return True

    @property
    def name(self):
        """Catalog name. can be edited"""
        return self._name

    @name.setter
    def name(self, new_cat_name):
        try:
            self._name = Stringifiable.cast_string(new_cat_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def datafiles(self):
        """Catalog :class:`~astrophysix.simdm.datafiles.Datafile` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._datafiles

    @property
    def catalog_fields(self):
        """Catalog :class:`~astrophysix.simdm.catalogs.field.CatalogField` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._fields

    @property
    def nobjects(self):
        """Returns the total number of objects in this catalog"""
        if len(self._fields) == 0:
            return 0
        return self._fields[0].nobjects

    @property
    def target_object(self):
        """Catalog associated :class:`~astrophysix.simdm.catalogs.targobj.TargetObject`."""
        return self._targobj

    @property
    def description(self):
        """Catalog description"""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def processing_services(self):
        """Catalog :class:`~astrophysix.simdm.catalogs.catalog.CatalogDataProcessingService` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._proc_services

    def _can_delete_object_property(self, obj_prop):
        """
        Checks if a target object property is not linked to any catalog field and can be safely deleted.
        Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        obj_prop: ``:class:~astrophysix.simdm.catalogs.targobj.ObjectProperty``
            target object property about to be deleted

        Returns
        -------
        o: str or None
        """
        # Checks that a given ObjectProperty instance is not associated to any field of this catalog
        for f in self._fields:
            if f.object_property is obj_prop:  # Reference identity, not equality ??? Should work
                return "{s!s} {field!s}".format(s=self, field=f)
        return None

    def _can_delete_catalog_field(self, cat_field):
        """
        Checks if a catalog field is not bound to any catalog data processing service and can be safely deleted.
        Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        cat_field: ``:class:~astrophysix.simdm.catalogs.field.CatalogField``
            catalog field about to be deleted

        Returns
        -------
        o: str or None
        """
        # Checks that a given CatalogField is not associated to any CatalogDataProcessingService in this catalog
        bound_fields = []
        for serv in self._proc_services:
            for field_binding in serv.catalog_field_bindings:
                if field_binding.catalog_field is cat_field:  # Reference identity, not equality ??? Should work
                    bound_fields.append("{bind!s}".format(bind=field_binding))

        if len(bound_fields) > 0:
            return bound_fields

        return None

    def _can_add_proc_service(self, process_service):
        """
        CatalogDataProcessingService addition validity check method. Verifies that the added data processing service
        contains only CatalogFieldBinding objects associated to a CatalogField currently listed in this catalog.
        Otherwise raises an AttributeError.

        Parameters
        ----------
        process_service: :class:`~astrophysix.simdm.services.process.CatalogDataProcessingService`
            new data processing service to bind to the Catalog
        """
        for field_binding in process_service.catalog_field_bindings:
            if field_binding.catalog_field not in self._fields:
                err_msg = "Cannot add catalog processing service. {f!s} is not a field of the " \
                          "{cat!s}.".format(f=field_binding.catalog_field, cat=self)
                log.error((err_msg))
                raise AttributeError(err_msg)

    def _can_add_field_binding_in_proc_service(self, field_binding):
        """
        CatalogFieldBinding addition validity check method. Verifies that the added catalog field binding is associated
        to a CatalogField that already belongs to this Catalog. Otherwise raises an AttributeError.

        Parameters
        ----------
        field_binding: :class:`~astrophysix.simdm.services.process.CatalogFieldBinding`
            new catalog field <=> data processing service binding to add to a data processing service within this
            Catalog.
        """
        if field_binding.catalog_field not in self._fields:
            err_msg = "Cannot add catalog field binding in this data processing service, {f!s} is not a " \
                      "field of the {cat!s}.".format(f=field_binding.catalog_field, cat=self)
            log.error((err_msg))
            raise AttributeError(err_msg)

    def _does_field_have_same_number_of_items(self, cat_field):
        """
        CatalogField addition validity check nethod. Verifies that the added field has the same number of items in it
        then other catalog fields, if there is any. Raises an AttributeError in case number of items differ.

        Parameters
        ----------
        cat_field: ``CatalogField``
            new catalog field to add to the Catalog
        """
        if self.nobjects == 0:  # No field yet in catalog
            return

        if cat_field.nobjects != self.nobjects:  # Number of catalog items differ => error
            err_msg = "Cannot add a catalog field with {n:d} item values in a catalog containing {cn:d} " \
                      "items.".format(n=cat_field.nobjects, cn=self.nobjects)
            log.error((err_msg))
            raise AttributeError(err_msg)

    def _does_field_property_is_target_obj_prop(self, cat_field):
        """
        CatalogField addition validity check method. Verifies that the added field property is one of the catalog's
        target object property. Otherwise raises an AttributeError.

        Parameters
        ----------
        cat_field: ``CatalogField``
            new catalog field to add to the Catalog
        """
        if cat_field.object_property not in self._targobj.object_properties:
            err_msg = "Cannot add catalog field. {p!s} is not a property of " \
                      "{tobj!s}.".format(tobj=self._targobj, p=cat_field.object_property)
            log.error((err_msg))
            raise AttributeError(err_msg)

    def to_pandas(self):
        """
        Convert a Catalog into a Pandas Dataframe object

        Returns
        -------
        df: pandas.DataFrame
            pandas DataFrame containing the catalog data
        """
        series_dict = {}
        for cat_field in self._fields:
            s = cat_field.to_pandas()
            series_dict[s.name] = s
        return pd.DataFrame(data=series_dict)

    def save_HDF5(self, catalog_fname=None, galactica_checks=False):
        """
        Save the Catalog into a HDF5 (\\*.h5) file

        *New in version 0.7.0*.

        Parameters
        ----------
        catalog_fname: :obj:`string`
            Catalog HDF5 filename.
        galactica_checks: :obj:`bool`
            Perform Galactica database validity checks and display warning in case of invalid content for upload on
            Galactica. Default False (quiet mode).
        """
        if catalog_fname is None:
            # No file path provided : should never happen
            err_msg = "No filename provided. Please provide a HDF5 filename to save the catalog."
            log.error(err_msg)
            raise AttributeError(err_msg)
        else:
            dest_catalog_h5fname = FileUtil.valid_filepath(catalog_fname, FileType.HDF5_FILE)

        # Move temporary file to the requested study HDF5 file path
        overwrite_catalog_file = False
        if os.path.exists(dest_catalog_h5fname):  # Delete first the already existing file
            while True:
                answer = input("File '{ch5f:s}' already exists. Overwrite? [Y/n] ".format(ch5f=dest_catalog_h5fname))
                if answer == "Y" or answer == "y":
                    overwrite_catalog_file = True
                    break
                else:
                    return

        # If Galactica database validity checks is enabled, perform a full check of the project before saving it.
        if galactica_checks:
            self.galactica_validity_check()

        # Save catalog into the temp. file
        temp_catalog_h5fname = FileUtil.new_temp_filepath("catalog")
        h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(temp_catalog_h5fname , mode=Hdf5StudyFileMode.NEW_WRITE)
        try:
            Hdf5StudyPersistent._hsp_write_object(h5group, "CATALOG", self, new_file=True, dry_run=False,
                                                  from_project=False)
            h5group.attrs["ObjectClass"] = "astrophysix.simdm.catalogs.Catalog"
            h5group.attrs["creation_time"] = DatetimeUtil.utc_to_timestamp(DatetimeUtil.utc_now())
        except Exception:
            raise
        finally:
            if close_when_done and h5f is not None:
                h5f.close()

        # Move temporary file to the requested study HDF5 file path
        if os.path.exists(dest_catalog_h5fname) and overwrite_catalog_file:  # Delete first the already existing file
            os.remove(dest_catalog_h5fname)
        shutil.copy(temp_catalog_h5fname, dest_catalog_h5fname)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "target_object",  "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Catalog object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Catalog into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Catalog, self)._hsp_write(h5group, **kwargs)

        # Write catalog name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write catalog description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write datafiles, if any defined
        self._hsp_write_object_list(h5group, "DATAFILES", self._datafiles, "datafile_", **kwargs)

        # Write target object
        if kwargs.get("from_project", False):  # Write target object UUID
            self._hsp_write_attribute(h5group, ('targob_uid', self._targobj.uid), **kwargs)
        else:  # Write complete target object description (full serialization)
            self._hsp_write_object(h5group, "TARGET_OBJECT", self._targobj, **kwargs)

        # Write catalog fields, if any defined
        self._hsp_write_object_list(h5group, "CATALOG_FIELDS", self._fields, "cat_field_", **kwargs)

        # Once catalog fields have been written, then write data processing services (and eventually their bindings to
        # catalog fields) bound to this catalog, if any defined.
        self._hsp_write_object_list(h5group, "PROC_SERVICES", self._proc_services, "proc_serv_", **kwargs)

        self._hsp_write_callback(str(self), **kwargs)

    @classmethod
    def load_HDF5(cls, catalog_file_path):
        """
        Loads Catalog from a HDF5 (\\*.h5) file

        *New in version 0.7.0*.

        Parameters
        ----------
        catalog_file_path: :obj:`string`
            Catalog HDF5 (existing) file path

        Returns
        -------
        study: :class:`~astrophysix.simdm.catalogs.catalog.Catalog`
            Catalog loaded from HDF5 file.
        """
        if not os.path.isfile(catalog_file_path):
            raise AttributeError("Cannot find file '{fname:s}'".format(fname=catalog_file_path))

        temp_loaded_file = FileUtil.new_temp_filepath("catalog")
        shutil.copy(catalog_file_path, temp_loaded_file)
        h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(temp_loaded_file,
                                                                        mode=Hdf5StudyFileMode.READ_ONLY)
        try:
            # Check that we are actually reading a Catalog object from a HDF5 file
            if "ObjectClass" not in h5group.attrs or h5group.attrs["ObjectClass"] != "astrophysix.simdm.catalogs.Catalog":
                err_msg = "HDF5 file does not contain a catalog !"
                log.error(err_msg)
                raise IOError(err_msg)

            # Read catalog creation/last modification times
            if "creation_time" not in h5group.attrs:
                err_msg = "Cannot find study creation time attributes in " \
                          "'{path!s}'.".format(path=h5group.name)
                log.error(err_msg)
                raise IOError(err_msg)

            if "CATALOG" not in h5group:
                err_msg = "Missing '/CATALOG' group in HDF5 catalog file."
                log.error(err_msg)
                raise IOError(err_msg)
            catalog_group = h5group["CATALOG"]
            cat = cls.hsp_load_from_h5(catalog_group, dependency_objdict={})
            cat._created = DatetimeUtil.utc_from_timestamp(h5group.attrs["creation_time"])
        except Exception:
            raise
        finally:
            if close_when_done and h5f is not None:
                h5f.close()

        return cat

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Catalog object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        catalog: ``Catalog``
            Read Catalog instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Catalog, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Try to read/find protocol
        try:
            targobj_uid = uuid.UUID(cls._hsp_read_attribute(h5group, "targob_uid", "target object UUID",
                                                            raise_error_if_not_found=True))

            # Search for already instantiated target object in dependency object dictionary
            if dependency_objdict is None:
                err_msg = "Cannot find any target object already instantiated in the project."
                log.error(err_msg)
                raise IOError(err_msg)

            # Get dictionary of target object instances of the corresponding class :
            if TargetObject.__name__ not in dependency_objdict:
                err_msg = "Cannot find any {cname:s} instance.".format(cname=TargetObject.__name__)
                log.error(err_msg)
                raise IOError(err_msg)

            # Find target object according to its UUID
            tobj_dict = dependency_objdict[TargetObject.__name__]
            if targobj_uid not in tobj_dict:
                err_msg = "Cannot find {cname:s} instance with uid {uid:s}.".format(cname=TargetObject.__name__,
                                                                                    uid=str(uid))
                log.error(err_msg)
                raise IOError(err_msg)

            tobj = tobj_dict[targobj_uid]
        except IOError:  # Target Object UUID not found in Catalog
            # Read target object info from "TARGET_OBJECT" subgroup
            tobj = TargetObject._hsp_read_object(h5group, "TARGET_OBJECT", "catalog target object",
                                                 dependency_objdict=dependency_objdict)

        # Read catalog name
        name = cls._hsp_read_attribute(h5group, 'name', "generic result name")

        # Create catalog object
        catalog = cls(uid=uid, target_object=tobj, name=name)

        # Build datafile list and add each datafile into catalog
        for df in Datafile._hsp_read_object_list(h5group, "DATAFILES", "datafile_", "catalog datafile",
                                                 dependency_objdict=dependency_objdict):
            catalog.datafiles.add(df)

        # Build catalog field list and add each field into catalog + build catalog field dictionary indexed by their
        # UUID
        if CatalogField.__name__ not in dependency_objdict:
            dependency_objdict[CatalogField.__name__] = {}
        cat_field_dict = dependency_objdict[CatalogField.__name__]
        for field in CatalogField._hsp_read_object_list(h5group, "CATALOG_FIELDS", "cat_field_", "catalog field",
                                                        dependency_objdict=dependency_objdict):
            catalog.catalog_fields.add(field)
            cat_field_dict[field.uid] = field

        # Read catalog description, if defined
        cat_descr = cls._hsp_read_attribute(h5group, 'description', "catalog description",
                                            raise_error_if_not_found=False)
        if cat_descr is not None:
            catalog.description = cat_descr

        # Read data processing services, if any defined
        if version >= 2 and "PROC_SERVICES" in h5group:
            for dps in CatalogDataProcessingService._hsp_read_object_list(h5group, "PROC_SERVICES", "proc_serv_",
                                                                          "catalog data processing service",
                                                                          dependency_objdict=dependency_objdict):
                catalog.processing_services.add(dps)

        return catalog

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check target object name
        if len(self._name) > 64:
            log.warning("{c!s} name is too long for Galactica (max. 64 characters).".format(c=self))

        # Check that the catalog contain some object
        if self.nobjects == 0:
            log.warning("{c!s} does not contain any object.".format(c=self))

        # Check catalog field and datafile list validity
        self._fields.galactica_validity_check(**kwargs)
        self._datafiles.galactica_validity_check(**kwargs)

        # Check data processing service list validity
        self._proc_services.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{cat_name:s}' catalog".format(cat_name=self._name)


__all__ = ["Catalog"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, int
import pytest
import logging
from astrophysix import units as U
from astrophysix.simdm import Project, ProjectCategory, SimulationStudy
from astrophysix.simdm.catalogs import Catalog, TargetObject
from astrophysix.simdm.experiment import Simulation
from astrophysix.simdm.protocol import SimulationCode
from astrophysix.simdm.results import GenericResult, Snapshot
from astrophysix.simdm.datafiles import Datafile
from astrophysix.simdm.services import DataProcessingService


class TestGenericResult(object):
    def test_generic_result_init(self):
        """
        Tests generic result instance initialisation
        """
        # Generic result initialisation
        res = GenericResult(name="Light-mass correlation", description="This is the paper major key result",
                            directory_path="/path/to/my/result")

        assert res.description == "This is the paper major key result"
        assert res.name == "Light-mass correlation"
        assert res.directory_path == "/path/to/my/result"

        # Test str() conversion
        assert str(res) == "'Light-mass correlation' generic result"

    def test_setting_generic_result_name_exc(self):
        """
        Tests that a generic result defined without name raises an exception
        """
        # Tests generic result defined without name
        with pytest.raises(AttributeError) as e:
            res_exc = GenericResult()  # No generic result name defined
        assert str(e.value) == "GenericResult 'name' attribute is not defined (mandatory)."

        res = GenericResult(name="My key result")

        # ------------- Tests invalid generic result name setting exception ---------------------- #
        empty_name_err = "GenericResult 'name' property is not a valid (non-empty) string."
        with pytest.raises(AttributeError) as e:
            res.name = -1
        assert str(e.value) == empty_name_err

        with pytest.raises(AttributeError) as e:
            res.name = ""
        assert str(e.value) == empty_name_err
        # ---------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        res.name = "My best result"

    def test_setting_generic_result_description(self):
        """
        Tests setting generic result description property
        """
        res = GenericResult(name="My super result")

        # -------------------------- Tests invalid generic result description setting exception ---------------------- #
        with pytest.raises(AttributeError) as e:
            res.description = 0
        assert str(e.value) == "GenericResult 'description' property is not a valid string."

        # Valid => should not raise any exception
        res.description = "My result description"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_generic_result_dirpath(self):
        """
        Tests setting generic result directory path property
        """
        res = GenericResult(name="My awesome key result")

        # --------------------- Tests invalid generic result directory path setting exception ------------------------ #
        with pytest.raises(AttributeError) as e:
            res.directory_path = (-1, ["Banana", "Apple"])
        assert str(e.value) == "GenericResult 'directory_path' property is not a valid string."

        # Valid => should not raise any exception
        res.directory_path = "/better/path/to/my/result"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_generic_result_equality(self):
        """
        Tests generic result rich comparison method GenericResult.__eq__()
        """
        res = GenericResult(name="Light-mass correlation", description="This is the paper major key result",
                            directory_path="/path/to/my/result")

        # Different UUID => not equals
        assert res != GenericResult(name=res.name, description=res.description, directory_path=res.directory_path)

        # Different name => not equals
        assert res != GenericResult(name="Name", description=res.description, directory_path=res.directory_path,
                                    uid=res.uid)

        # Different description => not equals
        assert res != GenericResult(name=res.name, description="Alt. description", directory_path=res.directory_path,
                                    uid=res.uid)

        # Different directory path => not equals
        assert res != GenericResult(name=res.name, description=res.description, directory_path="/alt/path/to/result",
                                    uid=res.uid)

        # Identical generic results
        r = GenericResult(name=res.name, description=res.description, directory_path=res.directory_path, uid=res.uid)
        assert res == r

        # Add datafiles
        df1 = res.datafiles.add(Datafile(name="My important datafile"))
        df2 = res.datafiles.add(Datafile(name="My important datafile (2)"))
        # Datafile list differ => not equals
        r.datafiles.add(df1)
        assert r != res
        # Identical generic results
        r.datafiles.add(df2)
        assert r == res

        # Add products in generic result
        tobj = TargetObject(name="Star")
        cat = res.catalogs.add(Catalog(tobj, name="My catalog"))
        # Catalog list differ => not equals
        assert r != res
        # Identical generic results
        r.catalogs.add(cat)
        assert r == res

    def test_generic_result_galactica_validity_checks(self, caplog):
        """
        Tests generic result Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        res = GenericResult(name="Light-mass correlation")

        # Result name too long
        res.name = "This is a way too long result name for a 'Light-mass correlation' basic result."
        res.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'This is a way too long result name for a 'Light-mass correlation' basic "
                                            "result.' generic result name is too long for Galactica (max. 64 "
                                            "characters).")

    def test_generic_result_hdf5_io(self, tmp_path):
        """
        Tests saving/loading GenericResult from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        res1 = GenericResult(name="Key result 1 !", description="My description", directory_path="/my/path/to/result")
        res2 = GenericResult(name="Key result 2 !", description="My description", directory_path="/path/to/result")
        tobj = TargetObject(name="My galaxy")
        res1.catalogs.add(Catalog(tobj, name="My catalog"))
        res2.catalogs.add(Catalog(tobj, name="My catalog #2"))
        res1.datafiles.add(Datafile(name="My datafile"))
        res2.datafiles.add(Datafile(name="My datafile (2)"))

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res1)
        simu.generic_results.add(res2)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare snapshots
        res1_loaded = simu_loaded.generic_results[res1.name]
        res2_loaded = simu_loaded.generic_results[res2.name]
        assert res1_loaded == res1
        assert res2_loaded == res2


class TestSnapshot(object):
    def test_snapshot_init(self):
        """
        Tests Snapshot instance initialisation
        """
        # Snapshot initialisation
        sn = Snapshot(name="First pericenter", time=(105.6, "Myr"), physical_size=(1.0e3, "pc"),
                      description="This is the snapshot of first pericenter", data_reference="OUTPUT_00245")

        assert sn.description == "This is the snapshot of first pericenter"
        assert sn.time[0] == 105.6
        assert sn.time[1] == U.Myr
        assert sn.physical_size[0] == 1000.0
        assert sn.physical_size[1] == U.pc
        assert sn.data_reference == "OUTPUT_00245"

        # Test str() conversion
        assert str(sn) == "'First pericenter' snapshot"

    def test_setting_snapshot_time_value_and_unit(self):
        """
        Tests setting snapshot time property
        """
        sn = Snapshot(name="My super snapshot")

        # ---------------------------- Tests invalid snapshot time setting exception --------------------------------- #
        # Length 3 tuple
        with pytest.raises(AttributeError) as e:
            sn.time = ("Apple", "Elephant", -3.14159)
        assert str(e.value) == "Snapshot 'time' property cannot be of length != 2"

        # Time unit => dict
        with pytest.raises(AttributeError) as e:
            sn.time = (102.6, {"my_unit": -4.3})
        assert str(e.value) == "Snapshot 'time' property must be defined with a valid (non-empty) time unit string."

        # "Empty time unit string
        with pytest.raises(AttributeError) as e:
            sn.time = (1.24, "")
        assert str(e.value) == "Snapshot 'time' property must be defined with a valid (non-empty) time unit string."

        # Unknwon time unit 'carots'
        with pytest.raises(AttributeError) as e:
            sn.time = (52.9, "carots")
        assert str(e.value) == "Snapshot 'time' property error : Unknown unit name 'carots'."

        # Time unit with physical type => mass instead of time
        with pytest.raises(AttributeError) as e:
            sn.time = (568.7, U.Msun)
        assert str(e.value) == "Error while setting Snapshot 'time' property : unit is not a valid time unit " \
                               "(physical type: 'mass')"

        # Invalid time value
        with pytest.raises(AttributeError) as e:
            sn.time = ("3.45.5", "Myr")
        assert str(e.value) == "Snapshot 'time' property must be set as a (time_float_value, time_unit) tuple."
        with pytest.raises(AttributeError) as e:
            sn.time = ([42359, "Goats"], "Myr")
        assert str(e.value) == "Snapshot 'time' property must be set as a (time_float_value, time_unit) tuple."

        # Valid => should not raise any exception
        sn.time = "0.256"
        sn.time = ("0.24", U.year)
        sn.time = ("0.45", "Myr")
        assert sn.time[1] == U.Myr
        sn.time = 4.46
        sn.time = (7.89e2, "Gyr")
        assert sn.time[1] == U.Gyr
        sn.time = (78.54, U.min)
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_snapshot_physsize_value_and_unit(self):
        """
        Tests setting snapshot physical size property
        """
        sn = Snapshot(name="My super snapshot")

        # ---------------------------- Tests invalid snapshot time setting exception --------------------------------- #
        # Length 3 tuple
        with pytest.raises(AttributeError) as e:
            sn.physical_size = ("Apple", "Elephant", -3.14159)
        assert str(e.value) == "Snapshot 'physical_size' property cannot be of length != 2"

        # Time unit => dict
        with pytest.raises(AttributeError) as e:
            sn.physical_size = (102.6, {"my_unit": -4.3})
        assert str(e.value) == "Snapshot 'physical_size' property must be defined with a valid (non-empty) length " \
                               "unit string."

        # "Empty time unit string
        with pytest.raises(AttributeError) as e:
            sn.physical_size = (1.24, "")
        assert str(e.value) == "Snapshot 'physical_size' property must be defined with a valid (non-empty) length " \
                               "unit string."

        # Unknwon time unit 'carots'
        with pytest.raises(AttributeError) as e:
            sn.physical_size = (52.9, "carots")
        assert str(e.value) == "Snapshot 'physical_size' property error : Unknown unit name 'carots'."

        # Length unit with physical type => energy instead of length
        with pytest.raises(AttributeError) as e:
            sn.physical_size = (568.7, U.J)
        assert str(e.value) == "Error while setting Snaphsot 'physical_size' property : unit is not a valid length " \
                               "unit (physical type: 'energy')"

        # Invalid physical size value
        with pytest.raises(AttributeError) as e:
            sn.physical_size = ("3.45.5", "kpc")
        assert str(e.value) == "Snapshot 'physical_size' property must be set as a (size_float_value, length_unit) " \
                               "tuple."
        with pytest.raises(AttributeError) as e:
            sn.physical_size = ([42359, "Goats"], "ly")
        assert str(e.value) == "Snapshot 'physical_size' property must be set as a (size_float_value, length_unit) " \
                               "tuple."

        # Valid => should not raise any exception
        sn.physical_size = "0.256"
        sn.physical_size = ("0.24", U.pc)
        sn.physical_size = ("0.45", "kpc")
        assert sn.physical_size[1] == U.kpc
        sn.physical_size = 4.46
        sn.physical_size = (7.89e2, "Mpc")
        assert sn.physical_size[1] == U.Mpc
        sn.physical_size = (78.54, U.ly)
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_snapshot_data_reference_exc(self):
        """
        Tests setting snapshot data_reference property
        """
        sn = Snapshot(name="My super snapshot")

        # ----------------------------- Tests invalid snapshot data_reference setting exception ---------------------- #
        with pytest.raises(AttributeError) as e:
            sn.data_reference = {'abcd': (0.1, None)}
        assert str(e.value) == "Snapshot 'data_reference' property is not a valid string."

        # Valid => should not raise any exception
        sn.data_reference = "OUTPUT_00034"

    def test_snapshot_equality(self):
        """
        Tests snapshot rich comparison method Snapshot.__eq__()
        """
        sn = Snapshot(name="My best snapshot ! Délire !", time=(542.1, U.Myr), physical_size=(250.0, U.kpc),
                      data_reference="output_00034")

        # Different UUID => not equals
        assert sn != Snapshot(name=sn.name, time=sn.time, physical_size=sn.physical_size,
                              data_reference=sn.data_reference)

        # Different time => not equals
        assert sn != Snapshot(name=sn.name, time=(542.1, U.kyr), physical_size=sn.physical_size, uid=sn.uid,
                              data_reference=sn.data_reference)

        # Different physical size => not equals
        assert sn != Snapshot(name=sn.name, time=sn.time, physical_size=(0.251, U.Mpc), uid=sn.uid,
                              data_reference=sn.data_reference)

        # Different data reference => not equals
        assert sn != Snapshot(name=sn.name, time=sn.time, physical_size=sn.physical_size, uid=sn.uid,
                              data_reference="other_output")

        # Identical snapshots
        s = Snapshot(name=sn.name, time=sn.time, physical_size=sn.physical_size, uid=sn.uid,
                     data_reference=sn.data_reference)
        assert s == sn

        # Compare bound data processing services
        dps = sn.processing_services.add(DataProcessingService(service_name="my_service", data_host="my_host"))
        assert s != sn
        s.processing_services.add(dps)
        assert s == sn

    def test_snapshot_galactica_validity_checks(self, caplog):
        """
        Tests snapshot Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        sn = Snapshot(name="Simple snapshot", time=(542.1, U.Myr), physical_size=(250.0, U.kpc),
                      data_reference="output_00038")

        # Check invalid time value
        sn.time = (1.2E6, U.year)
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Simple snapshot' snapshot time value must be defined in the range "
                                            "[-10000.0, 10000.0].")
        sn.time = (542.1, U.Myr)  # => Ok

        # Check invalid physical size value
        sn.physical_size = (2.5E8, U.au)
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Simple snapshot' snapshot physical size value must be defined in the "
                                            "range [0.0, 10000.0].")
        sn.physical_size = (250.0, U.kpc)  # => Ok

        # Check data reference
        sn.data_reference = "data_reference whi is way too long for this simple snapshot at t=542.1 Myr"
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Simple snapshot' snapshot data reference is too long (max. 64 "
                                            "characters).")
        sn.data_reference = "output_00038"  # => Ok

        # Check data reference is not empty if snapshot is bound to a data processing service
        sn.data_reference = ""
        sn.processing_services.add(DataProcessingService(service_name="ray_traced_map", data_host="anais"))
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Simple snapshot' snapshot data reference is not defined while data "
                                            "processing services are bound to it.")

    def test_snapshot_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Snapshot from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        sn = Snapshot(name="My best snapshot ! Délire !", description="My description", time=542.1,
                      physical_size=(250.0, U.kpc), directory_path="/path/to/snapshot1", data_reference="OUTPUT_00001")
        sn2 = Snapshot(name="My best snapshot ! Délire ! (2)", description="My description", time=(542.1, U.Myr),
                       physical_size=255.48, directory_path="/path/to/snapshot2", data_reference="OUTPUT_00002")
        sn.datafiles.add(Datafile(name="Datafile 1"))
        sn2.datafiles.add(Datafile(name="Datafile 2"))

        # Add data processing services
        sn.processing_services.add((DataProcessingService(service_name="slice_map", data_host="my_cluster")))
        sn.processing_services.add((DataProcessingService(service_name="ppv_cube", data_host="my_GPU_cluster")))

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.snapshots.add(sn)
        simu.snapshots.add(sn2)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare snapshots
        sn_loaded = simu_loaded.snapshots[sn.name]
        sn_loaded2 = simu_loaded.snapshots[sn2.name]
        assert sn_loaded == sn
        assert sn_loaded2 == sn2


class TestDataProcessingService(object):
    def test_proc_service_init(self):
        """
        Tests DataProcessingService instance initialisation
        """
        # Data processing service initialisation
        service = DataProcessingService(service_name="ray_traced_map", data_host="lab_cluster")

        assert service.service_name == "ray_traced_map"
        assert service.data_host == "lab_cluster"
        assert service.hosted_service == "ray_traced_map @ lab_cluster"

        # Test str() conversion
        assert str(service) == "'ray_traced_map' data processing service (host: lab_cluster)"

    def test_setting_process_service_name_exc(self):
        """
        Tests that a data processing service defined without a name raises an exception
        """
        # Data processing service defined without service name
        with pytest.raises(AttributeError) as e:
            dps_exc = DataProcessingService(data_host="lab_GPUCLUST")  # No service name defined
        assert str(e.value) == "Data processing service 'service_name' attribute is not defined (mandatory)."

        # Tests data processing service name property initialisation exception
        invalid_dataprocserv_name_err = "Data processing service 'service_name' property is not a valid " \
                                        "\\(non empty\\) string."
        with pytest.raises(AttributeError, match=invalid_dataprocserv_name_err):
            dps_exc = DataProcessingService(service_name=-2.4, data_host="lab_GPUCLUST")  # Invalid data processing service name

        with pytest.raises(AttributeError, match=invalid_dataprocserv_name_err):
            dps_exc = DataProcessingService(service_name="", data_host="lab_GPUCLUST")  # Invalid data processing service name (empty string)

        dps = DataProcessingService(service_name="my_custom_Service", data_host="lab_GPUCLUST")
        with pytest.raises(AttributeError, match=invalid_dataprocserv_name_err):
            dps.service_name = -1  # Invalid data processing service name

    def test_setting_process_service_data_host(self):
        """
        Tests setting data processing service host server name property
        """
        # Data processing service defined without data host name
        with pytest.raises(AttributeError) as e:
            dps_exc = DataProcessingService(service_name="my_custom_service")  # No data host name defined
        assert str(e.value) == "Data processing service 'data_host' attribute is not defined (mandatory)."

        # Tests data processing service data host server property initialisation exception
        invalid_dataprocserv_ndata_host_err = "Data processing service 'data_host' property is not a valid " \
                                              "\\(non empty\\) string."
        with pytest.raises(AttributeError, match=invalid_dataprocserv_ndata_host_err):
            dps_exc = DataProcessingService(service_name="my_custom_service", data_host=-2.4)  # Invalid data processing service name

        with pytest.raises(AttributeError, match=invalid_dataprocserv_ndata_host_err):
            dps_exc = DataProcessingService(service_name="my_custom_service", data_host="")  # Invalid data processing service data host name (empty string)

        dps = DataProcessingService(service_name="my_custom_service", data_host="lab_gpu_cluster")
        with pytest.raises(AttributeError, match=invalid_dataprocserv_ndata_host_err):
            dps.data_host = {"a": 245.1}  # Invalid data processing service data host name

    def test_process_service_comparison(self):
        """
        Tests rich comparison method DataProcessingService.__eq__()
        """
        dps = DataProcessingService(service_name="my_service", data_host="my_workstation")

        # Different UUID => not equals
        assert dps != DataProcessingService(service_name=dps.service_name, data_host=dps.data_host)

        # Different service name => not equals
        assert dps != DataProcessingService(service_name="other_service", uid=dps.uid, data_host=dps.data_host)

        # Different data host => not equals
        assert dps != DataProcessingService(service_name=dps.service_name, uid=dps.uid, data_host="other_host")

        # Identical data processing service
        assert dps == DataProcessingService(service_name=dps.service_name, uid=dps.uid, data_host=dps.data_host)

    def test_data_proc_service_hdf5_io(self, tmp_path):
        """
        Tests saving/loading DataProcessingService from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        sn = Snapshot(name="My best snapshot ! Délire !")
        # Add data processing services
        dps1 = sn.processing_services.add((DataProcessingService(service_name="slice_map", data_host="my_cluster")))
        dps2 = sn.processing_services.add((DataProcessingService(service_name="ppv_cube", data_host="my_GPU_cluster")))

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.snapshots.add(sn)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]
        sn_loaded = simu_loaded.snapshots[sn.name]

        # Compare services
        dps1_loaded = sn_loaded.processing_services[dps1.hosted_service]
        dps2_loaded = sn_loaded.processing_services[dps2.hosted_service]
        assert dps1 == dps1_loaded
        assert dps2 == dps2_loaded

    def test_data_proc_service_galactica_validity_checks(self, caplog):
        """
        Tests data processing service Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        sn = Snapshot(name="My best snapshot.")

        # Add data processing services
        dps1 = sn.processing_services.add((DataProcessingService(service_name="My_super_service_that_is_way_too_long",
                                                                 data_host="my_cluster")))
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "Data processing service name 'My_super_service_that_is_way_too_long' is "
                                            "too long for Galactica (max. 32 characters).")

        dps1.service_name = "slice_map"
        dps1.data_host = "My_long_host_name"
        sn.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "Data processing service host name 'My_long_host_name' is too long for "
                                            "Galactica (max. 16 characters).")


__all__ = ["TestSnapshot", "TestGenericResult", "TestDataProcessingService"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@startuml

Protocol <|-- SimulationCode
Protocol <|-- PostProcessingCode

Protocol "1" *-- "many" InputParameter : contains
Protocol "1" *-- "many" Algorithm : contains
SimulationCode "1" *-- "many" PhysicalProcess : contains

class Protocol {
   input_parameters : ObjectList
   algorithms : ObjectList
}

class SimulationCode {
    physical_processes : ObjectList
}

@enduml
"""
from __future__ import absolute_import, unicode_literals
from .input_parameters import InputParameter
from .algorithm import Algorithm, AlgoType
from .physics import Physics, PhysicalProcess
from .base import SimulationCode, PostProcessingCode


__all__ = ["InputParameter", "AlgoType", "Algorithm", "Physics", "PhysicalProcess",
           "PostProcessingCode", "SimulationCode"]
